package MkPPD;

use strict;

our $VERSION = '3.00';

use Config qw(%Config);
use ActiveState::Handy qw(xml_esc);

# You should provide:
# - name
# - version
# - author
# - abstract
sub new {
    my($class, %hash) = @_;
    die "Missing name" unless $hash{name};
    die "Missing verson" unless $hash{version};
    
    if ($hash{version} =~ tr/.// <= 1) {
	# Rewrite 1.203 as 1.2.3
	$hash{version} =~ s/^(\d+\.\d)(\d\d)$/
	                    join(".", $1,int($2))
                           /e || die "Bad version number $hash{version}";

	if (exists $ENV{PMX_BUILD}) {
	    die "Bad \$PMX_BUILD value" unless $ENV{PMX_BUILD} =~ /^\d+$/;
	    $hash{version} .= ".$ENV{PMX_BUILD}";
	}
	else {
	    die "No \$PMX_BUILD value defined";
	}
    }

    $hash{ppd_version} = _ppd_version($hash{version});

    if ($hash{implementations}) {
	for my $impl (@{$hash{implementations}}) {
	    $impl->{fullname} ||= do {
		my $n = "$hash{name}-$hash{version}";
		if (my $a = $impl->{architecture}) {
		    $a =~ s/-(thread|multi)//g;
		    $n .= "-$a";
		}
		$n;
	    };
	}
    }
    else {
	my %impl;
	my $fullname = "$hash{name}-$hash{version}";
	if ($hash{arch}) {
	    my $archname = $Config{archname};
	    (my $platform = $archname) =~ s/-(thread|multi)//g;
	    $fullname .= "-$platform";
	    $impl{architecture} = $archname;
	    $impl{os} = $^O;
	}
	$impl{fullname} = $fullname;
	$hash{implementations} = [\%impl];
    }

    $hash{author} ||= "Sophos Plc.";
    $hash{abstract} ||= "";

    if (my $script = $hash{install}) {
	if (ref($script) || ref(\$script) eq "GLOB") {
	    $script = do { local $/; <$script> };
	}
	
	# expand some variables too
	$script =~ s/%%([A-Z_]+)%%/$hash{lc($1)}/g;
	$hash{install} = $script;

	$hash{install_exec} ||= "PPM_PERL";
    }

    bless \%hash, $class;
    return \%hash;
}

sub finish {
    my $self = shift;
    $self->write;
    for my $t ($self->tarballs) {
	$t->finish;
    }
}

sub tarball {
    my $self = shift;
    my $n = shift;
    $n = 0 unless defined $n;
    my $name = ($n =~ /^\d+$/) ? $self->{implementations}[$n]{fullname} : $n;
    die "Tarball $n does not have a name" unless $name;

    $self->{tarball}{$name} ||= do {
        MkPPD::Tarball->new($name);
    };
}

sub tarballs {
    my $self = shift;
    return values %{$self->{tarball}};
}

sub write {
    my $self = shift;
    my $ppd = "$self->{name}.ppd";
    open(my $fh, ">", $ppd) || die "Can't create $ppd: $!";
    print $fh <<EOT;
<SOFTPKG NAME="$self->{name}" VERSION="$self->{ppd_version}">
<TITLE>$self->{name}</TITLE>
<ABSTRACT>$self->{abstract}</ABSTRACT>
<AUTHOR>$self->{author}</AUTHOR>
EOT

    for my $impl (@{$self->{implementations}}) {
	print $fh "\n<IMPLEMENTATION>\n";
	for (qw(os architecture)) {
	    print $fh qq(<\U$_\E NAME="$impl->{$_}"/>\n)
		if $impl->{$_};
	}

	if (my $req = $self->{requires}) {
	    $req = [$req] unless ref($req);
	    for (@$req) {
		my $v = 0;
		my $n = $_;
		($n, $v) = ($1, $2) if /^(.*)-(\d+(\.\d+)*)$/;
		$v = _ppd_version($v);
		print $fh qq(<DEPENDENCY NAME="$n" VERSION="$v"/>\n);
	    }
	}

	my $tarball = $self->tarball($impl->{fullname});
	my $tarname = $tarball->name;
	print $fh qq(<CODEBASE HREF="$tarname"/>\n);
	if (my $script = $self->{install}) {
	    unless ($self->{inline_scripts}) {
		$tarball->write_file("install.pl", $script);
		$script = <<'EOT';
do './install.pl' || die "Install script did not succeed!\n$@$!\n";
EOT
            }

	    # ppm does not accept encoded tabs in the text
	    require Text::Tabs;
	    $script = join("", "BEGIN { unlink __FILE__ } # Old ppm workaround\n",
		               map Text::Tabs::expand($_), split(/^/, $script));

	    print $fh qq(\n<INSTALL EXEC="$self->{install_exec}">\n);
	    print $fh xml_esc($script);
	    print $fh qq(</INSTALL>\n\n);
	}

	print $fh "</IMPLEMENTATION>\n";
    }

    print $fh "\n</SOFTPKG>\n";

    close($fh) || die "Could not write $ppd: $!";
    print "# wrote $ppd\n";

    return 1;
}

sub _ppd_version {
    my $version = shift;
    my $count = ((my $ppd_version = $version) =~ s/\./,/g);
    $ppd_version =~ s/(^|,)0+(\d)/$1$2/g;
    $ppd_version .= ",0" x (3 - $count) if $count < 3;
    $ppd_version;
}

sub parse_requires {
    my @reqs;
    if (open(my $req_fh, "REQUIRES")) {
        my $badline = 0;
        while (<$req_fh>) {
            next if /^\s*\#/;
            next if /^\s*$/;
            if (/^\w+(-\w+)*-(\d+(\.\d+)*)$/) {
                chomp;
                push(@reqs, $_);
            }
            else {
                warn "Bad REQUIRES line: $_";
                ++$badline;
            }
        }
        die <<'EOT' if $badline;
All packages in REQUIRES must have a version number!
Use Module-0 if no particular version is needed.
EOT
    }
    return @reqs;
}

package MkPPD::Tarball;

use ActiveState::RelocateTree qw(move_tree);
use ActiveState::Handy qw(run);
use File::Path qw(mkpath rmtree);
use File::Find;

my $tarball_count = 0;

sub new {
    my($class, $basename) = @_;
    $tarball_count++;
    my $tmpdir = "tarball-$tarball_count-" . sprintf("%x", $$);
    die "Will not clobber $tmpdir" if -e $tmpdir;
    mkpath($tmpdir) or die "Can't mkpath $tmpdir: $!";

    my $self = bless {
	basename => $basename,
	tmpdir => $tmpdir,
    }, $class;

    return $self;
}

sub stat {
    my ($self, $file) = @_;
    return stat("$self->{tmpdir}/$file");
}

sub add {
    my $self = shift;
    my $from = shift;
    my $to = @_ ? shift : $from;
    die if $self->{finished};

    $to =~ s,^/,,;

    if (-d $from) {
	unless (-d "$self->{tmpdir}/$to") {
	    mkpath("$self->{tmpdir}/$to", 1) or 
		die "Can't mkpath $self->{tmpdir}/$to: $!";
	}
	move_tree($from, "$self->{tmpdir}/$to", 0, 1);
    }
    # If the link just points to a sibling, preserve it as a symlink
    elsif (-l $from and ((my $l = readlink $from) !~ m#^/#)) {
	die "'$to' already exists in tarball" if -e "$self->{tmpdir}/$to";
	if ($to =~ m#/#) {
	    (my $todir = $to) =~ s#/[^/]+\z## || die;
	    unless (-d "$self->{tmpdir}/$todir") {
		mkpath("$self->{tmpdir}/$todir", 1) or 
		    die "Can't mkpath $self->{tmpdir}/$todir: $!";
	    }
	}
	print "ln -s $self->{tmpdir}/$to $l\n";
	symlink($l, "$self->{tmpdir}/$to")
	    or die "can't create symlink $self->{tmpdir}/$to -> $l: $!";
    }
    else {
	die "'$to' already exists in tarball" if -e "$self->{tmpdir}/$to";
	if ($to =~ m,/,) {
	    (my $todir = $to) =~ s,/[^/]+\z,, || die;
	    unless (-d "$self->{tmpdir}/$todir") {
		mkpath("$self->{tmpdir}/$todir", 1) or 
		    die "Can't mkpath $self->{tmpdir}/$todir: $!";
	    }
	}
	run("cp $from $self->{tmpdir}/$to");
    }
}

sub add_blib {
    my $self = shift;
    $self->add("blib");
}

sub add_fake_blib {
    my $self = shift;
    die if $self->{finished};
    unless (-d "$self->{tmpdir}/blib/arch") {
	mkpath("$self->{tmpdir}/blib/arch") or 
	    die "Can't mkpath $self->{tmpdir}/blib/arch: $!";
    }
}

sub mv {
    my($self, $from, $to) = @_;
    die if $self->{finished};
    run("mv $self->{tmpdir}/$from $self->{tmpdir}/$to");
}

sub cp {
    my($self, $from, $to) = @_;
    die if $self->{finished};
    my $r = (-d "$self->{tmpdir}/$from") ? "r" : "";
    run("cp -r $self->{tmpdir}/$from $self->{tmpdir}/$to");
}

sub mkdir {
    my($self, $dir) = @_;
    die if $self->{finished};
    unless (-d "$self->{tmpdir}/$dir") {
	mkpath("$self->{tmpdir}/$dir") or 
	    die "Can't mkpath $self->{tmpdir}/$dir: $!";
    }
}

sub rm {
    my($self, $file) = @_;
    die if $self->{finished};
    rmtree("$self->{tmpdir}/$file") or 
        die "Can't rmtree $self->{tmpdir}/$file: $!";
}

sub write_file {
    my ($self, $file, $content, $mode) = @_;
    $mode ||= '>';
    $file = "$self->{tmpdir}/$file";
    chmod(0644, $file);
    open(my $f, "$mode", "$file") || die "Can't create $file: $!";
    binmode($f);
    print $f $content;
    close($f) || die "Can't write $file: $!";
}

# $match is an optional parameter. Use it to specify a custom perl path
# that will be replaced. See the reloc_perl -h documentation for more
# details.
sub reloc {
    my($self, $f, $match) = @_;
    $match = "" unless defined $match;
    require ActiveState::RelocateTree;
    ActiveState::RelocateTree::relocate(
	to => "$self->{tmpdir}/$f",
	inplace => 1,
	replace => ActiveState::RelocateTree::spongedir('ppm'),
        ($match ? (from => $match) : ()),
    );
}

sub name {
    my $self = shift;
    return $self->{basename} . ".tar.gz";
}

sub finish {
    my $self = shift;
    die if $self->{finished};
    my $tmpdir = $self->{tmpdir};
    my $basename = $self->{basename};
    # look for filenames longer than 100 characters
    my $max_len = 100 + length("$tmpdir/");
    find( sub { die "filename $File::Find::name is too long!" 
                    if length($File::Find::name) > $max_len },
          $tmpdir
        );
    run("cd $tmpdir && gtar --owner root --group bin -cf ../$basename.tar .");
    rmtree($tmpdir) or die "Can't rmtree $tmpdir: $!";
    run("gzip -f9 $basename.tar");
    $self->{finished}++;
}

sub DESTROY {
    my $self = shift;
    return if  $self->{finished};
    rmtree($self->{tmpdir}) or die "Can't rmtree $self->{tmpdir}: $!";
}

1;
